from rest_framework.pagination import PageNumberPagination as _PageNumberPagination, \
    LimitOffsetPagination as _LimitOffsetPagination, CursorPagination as _CursorPagination, _positive_int

from djackal.settings import djackal_settings


class BasePagination:
    def get_paginated_meta(self):
        raise NotImplementedError('get_paginated_meta() must be implemented.')


class PageNumberPagination(BasePagination, _PageNumberPagination):
    page_size = djackal_settings.PAGE_SIZE
    page_size_query_param = 'page_size'
    max_page_size = djackal_settings.MAX_PAGE_SIZE

    def get_paginated_meta(self):
        return {
            'count': self.page.paginator.count,
            'page': self.page.number,
        }


class LimitOffsetPagination(BasePagination, _LimitOffsetPagination):
    def get_paginated_meta(self):
        return {
            'count': self.count,
            'previous': self.get_previous_link(),
            'next': self.get_next_link(),
        }


class CursorPagination(BasePagination, _CursorPagination):
    def get_paginated_meta(self):
        return {
            'previous': self.get_previous_link(),
            'next': self.get_next_link(),
        }


class NoCountPagination(PageNumberPagination):
    last_page_strings = None

    def get_page_number(self, request):
        if self.page_size_query_param:
            try:
                return _positive_int(
                    request.query_params[self.page_query_param],
                )
            except (KeyError, ValueError):
                pass

        return 1

    def paginate_queryset(self, queryset, request, view=None):
        page_size = self.get_page_size(request)
        page_number = self.get_page_number(request)

        offset = page_size * (page_number - 1)

        # Get one extra element to check if there is a "next" page
        q = list(queryset[offset: offset + page_size + 1])
        self.count = offset + len(q) if len(q) else offset - 1
        self.has_next = False

        if len(q) > page_size:
            self.has_next = True
            q.pop()

        self.request = request
        return q

    def get_paginated_meta(self):
        return {
            'has_next': self.has_next
        }



import http
import random

from rest_framework.test import APIRequestFactory

from djackal.erra import Erra
from djackal.exceptions import DjackalAPIException, NotFound, Forbidden, BadRequest, InternalServer, ErraException, \
    Unauthorized, NotAllowed
from djackal.tests import DjackalAPITestCase
from djackal.views.base import DjackalAPIView


class TestErra(Erra):
    TEST_ERRA = 'test_erra_message'


TEST_BAD_REQUEST_CODE = 'TEST_BAD_REQUEST'
TEST_UNAUTHORIZED_CODE = 'TEST_UNAUTHORIZED'
TEST_FORBIDDEN_CODE = 'TEST_FORBIDDEN'
TEST_NOT_ALLOWED_CODE = 'TEST_NOT_ALLOWED'
TEST_INTERNAL_SERVER_CODE = 'TEST_INTERNAL_SERVER'

factory = APIRequestFactory()


class TestException(DjackalAPIException):
    default_status_code = 501


class ExceptionAPI(DjackalAPIView):
    def post(self, request):
        kind = request.data['kind']
        status_code = request.data.get('status_code')

        if kind == 'NotFound':
            raise NotFound(test=True)
        elif kind == 'BadRequest':
            raise BadRequest(code=TEST_BAD_REQUEST_CODE, test=True)
        elif kind == 'Unauthorized':
            raise Unauthorized(code=TEST_UNAUTHORIZED_CODE, test=True)
        elif kind == 'Forbidden':
            raise Forbidden(code=TEST_FORBIDDEN_CODE, test=True)
        elif kind == 'NotAllowed':
            raise NotAllowed(code=TEST_NOT_ALLOWED_CODE, test=True)
        elif kind == 'InternalServer':
            raise InternalServer(code=TEST_INTERNAL_SERVER_CODE, test=True)
        elif kind == 'Erra':
            raise ErraException(erra=TestErra.TEST_ERRA, test=True, status_code=status_code)

    def get(self, request):
        raise TestException()


class ExceptionTest(DjackalAPITestCase):
    def setUp(self):
        self.view = ExceptionAPI.as_view()

    def test_exception_handle(self):
        request = factory.get('/')
        response = self.view(request)
        self.assertStatusCode(501, response)

    def _check_sub_erra_exception(self, kind, code, status_code):
        request = factory.post('/', {'kind': kind})
        response = self.view(request)
        self.assertStatusCode(status_code, response)
        result = response.data
        self.assertEqual(result['test'], True)
        self.assertEqual(result['code'], code)

    def test_exception_response(self):
        request = factory.post('/', {'kind': 'NotFound'})
        response = self.view(request)
        self.assertStatusCode(404, response)

        result = response.data
        self.assertEqual(result['test'], True)
        self.assertEqual(result['code'], 'NOT_FOUND')

        self._check_sub_erra_exception('BadRequest', TEST_BAD_REQUEST_CODE, 400)
        self._check_sub_erra_exception('Unauthorized', TEST_UNAUTHORIZED_CODE, 401)
        self._check_sub_erra_exception('Forbidden', TEST_FORBIDDEN_CODE, 403)
        self._check_sub_erra_exception('NotAllowed', TEST_NOT_ALLOWED_CODE, 405)

        request = factory.post('/', {'kind': 'Erra'})
        response = self.view(request)
        self.assertStatusCode(500, response)
        result = response.data
        self.assertEqual(result['test'], True)
        self.assertEqual(result['code'], TestErra.TEST_ERRA.code)
        self.assertEqual(result['message'], TestErra.TEST_ERRA.message)

    def test_exception_status_code(self):
        request = factory.post('/', {'kind': 'Erra'})
        response = self.view(request)
        self.assertStatusCode(500, response)
        code = random.choice(list(http.HTTPStatus))

        request = factory.post('/', {'kind': 'Erra', 'status_code': code.value})
        response = self.view(request)
        self.assertStatusCode(code.value, response)



from django.db import models

from djackal.fields import JSONField
from djackal.tests import DjackalTransactionTestCase


class FieldTestModel(models.Model):
    json = JSONField(default=dict)


class FieldTest(DjackalTransactionTestCase):
    def test_json_field(self):
        test_dict = {'test_key': 'test_value'}

        tobj = FieldTestModel(json=test_dict)
        tobj.save()
        assert tobj.json == test_dict
        tobj = FieldTestModel.objects.first()
        assert tobj.json == test_dict

        tobj.json['test_key2'] = 'test_value2'
        tobj.save()

        assert tobj.json['test_key2'] == 'test_value2'



from hyena import Hyena
from puty import purify
from rest_framework.response import Response
from rest_framework.views import APIView

from djackal.exceptions import NotFound
from djackal.filters import DefaultFilterFunc
from djackal.settings import djackal_settings
from djackal.shortcuts import gen_q
from djackal.utils import value_mapper, isiter


class FilterMixin:
    lookup_map = {}
    filter_map = {}
    search_map = {}
    extra_map = {}
    ordering_map = {}

    search_keyword_key = 'search_keyword'
    search_type_key = 'search_type'
    ordering_key = 'ordering'
    ordering_default = None

    custom_action_prefix = '@'

    user_field = None
    bind_user_field = None
    filter_func_class = DefaultFilterFunc

    def get_filter_func_instance(self):
        return self.filter_func_class()

    def filter_func_action(self, name, value):
        instance = self.get_filter_func_instance()
        return instance.action(name, value)

    def get_lookup_map(self, **additional):
        d = self.lookup_map or dict()
        return {**d, **additional}

    def get_filter_map(self, **additional):
        d = self.filter_map or dict()
        return {**d, **additional}

    def get_search_map(self, **additional):
        d = self.search_map or dict()
        return {**d, **additional}

    def get_extra_map(self, **additional):
        d = self.extra_map or dict()
        return {**d, **additional}

    def get_ordering_map(self, **additional):
        d = self.ordering_map or dict()
        return {**d, **additional}

    def get_user_field(self):
        return self.user_field

    def filter_by_lookup_map(self, queryset, lookup_map=None):
        if lookup_map is None:
            lookup_map = self.get_lookup_map()

        mapped = value_mapper(lookup_map, self.kwargs)
        return queryset.filter(**mapped)

    def filter_by_user(self, queryset, user_field=None):
        if user_field is None:
            user_field = self.get_user_field()

        if not self.has_auth() or not user_field:
            return queryset
        return queryset.filter(**{user_field: self.request.user})

    def filter_by_extra_map(self, queryset, extra_map=None):
        if extra_map is None:
            extra_map = self.get_extra_map()
        return queryset.filter(**extra_map)

    def filter_by_filter_map(self, queryset, filter_map=None):
        params = self.request.query_params
        if filter_map is None:
            filter_map = self.get_filter_map()

        for map_key, map_value in filter_map.items():
            if map_key.endswith('[]') and hasattr(params, 'getlist'):
                value = params.getlist(map_key)
            else:
                value = params.get(map_key)
            if value in [None, '', []]:
                continue

            split_key = map_key.split(':')
            if len(split_key) == 2:
                value = self.filter_func_action(split_key[1], value)

            if map_value.startswith(self.custom_action_prefix):
                keyword = map_value.replace(self.custom_action_prefix, '')
                queryset = self.filter_by_filter_action(queryset, keyword, value)
                continue
            if isiter(map_value):
                queryset = queryset.filter(*gen_q(value, *map_value))
            else:
                queryset = queryset.filter(**{map_value: value})
        return queryset

    def filter_by_search_map(self, queryset, search_map=None):
        if search_map is None:
            search_map = self.get_search_map()

        params = self.request.query_params
        search_type = params.get(self.search_type_key)
        search_keyword = params.get(self.search_keyword_key)
        map_value = search_map.get(search_type)

        if not search_keyword or not map_value:
            return queryset

        if map_value.startswith(self.custom_action_prefix):
            keyword = map_value.replace(self.custom_action_prefix, '')
            return self.filter_by_search_action(queryset, keyword, search_keyword)
        if isiter(map_value):
            return queryset.filter(gen_q(search_keyword, *map_value))
        else:
            return queryset.filter(**{map_value: search_keyword})

    def filter_by_ordering(self, queryset, order_map=None):
        if order_map is None:
            order_map = self.get_ordering_map()

        params = self.request.query_params
        param_value = params.get(self.ordering_key)

        if order_map is not None:
            order_value = order_map.get(param_value)
        else:
            order_value = param_value

        if not order_value:
            if not self.ordering_default:
                return queryset
            order_value = self.ordering_default

        if order_value.start_with(self.custom_action_prefix):
            keyword = order_value.replace(self.custom_action_prefix, '')
            return self.filter_by_order_action(queryset, keyword)

        order_by = order_value.split(',')
        return queryset.order_by(*order_by)

    def filter_by_filter_action(self, queryset, keyword, value):
        return queryset

    def filter_by_search_action(self, queryset, keyword, value):
        return queryset

    def filter_by_order_action(self, queryset, keyword):
        return queryset

    def get_filtered_queryset(self, queryset=None):
        if queryset is None:
            queryset = self.get_queryset()

        queryset = self.filter_by_user(queryset)
        queryset = self.filter_by_lookup_map(queryset)
        queryset = self.filter_by_extra_map(queryset)
        queryset = self.filter_by_filter_map(queryset)
        queryset = self.filter_by_search_map(queryset)
        queryset = self.filter_by_ordering(queryset)

        return queryset

    def get_object(self, queryset=None):
        if queryset is None:
            queryset = self.get_queryset()

        queryset = self.filter_by_user(queryset)
        queryset = self.filter_by_lookup_map(queryset)
        queryset = self.filter_by_extra_map(queryset)

        obj = queryset.first()
        if obj is None:
            raise NotFound(model=queryset.model)
        self.check_object_permissions(request=self.request, obj=obj)
        return obj


class PageMixin:
    pagination_class = djackal_settings.DEFAULT_PAGINATION_CLASS
    paging = False

    @property
    def paginator(self):
        if not hasattr(self, '_paginator'):
            if not self.paging:
                self._paginator = None
            else:
                if self.pagination_class is None:
                    self._paginator = None
                else:
                    self._paginator = self.pagination_class()
        return self._paginator

    def get_paginate_queryset(self, queryset):
        if self.paginator is None:
            return None
        return self.paginator.paginate_queryset(queryset, self.request, view=self)

    def get_paginated_meta(self):
        return self.paginator.get_paginated_meta()


class BaseDjackalAPIView(APIView):
    default_permission_classes = ()
    default_authentication_classes = ()

    result_root = 'result'
    result_meta = 'meta'

    required_auth = False

    def get_client_ip(self, request):
        x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
        if x_forwarded_for:
            ip = x_forwarded_for.split(',')[0]
        else:
            ip = request.META.get('REMOTE_ADDR')
        return ip

    def get_default_exception_handler(self):
        return djackal_settings.EXCEPTION_HANDLER

    def get_authentication_classes(self):
        if self.authentication_classes:
            return (
                *self.default_authentication_classes,
                *self.authentication_classes
            )
        return self.default_authentication_classes

    def get_authenticators(self):
        return [auth() for auth in self.get_authentication_classes()]

    def get_permission_classes(self):
        if self.permission_classes:
            return (
                *self.default_permission_classes,
                *self.permission_classes
            )
        return self.default_permission_classes

    def get_permissions(self):
        return [permission() for permission in self.get_permission_classes()]

    def pre_check_object_permissions(self, request, obj):
        pass

    def pre_check_permissions(self, request):
        pass

    def pre_handle_exception(self, exc):
        pass

    def pre_method_call(self, request, *args, **kwargs):
        pass

    def post_check_object_permissions(self, request, obj):
        pass

    def post_check_permissions(self, request):
        pass

    def post_method_call(self, request, response, *args, **kwargs):
        pass

    def check_permissions(self, request):
        self.pre_check_permissions(request)
        super().check_permissions(request)
        self.post_check_permissions(request)

    def check_object_permissions(self, request, obj):
        self.pre_check_object_permissions(request, obj)
        super().check_object_permissions(request, obj)
        self.post_check_object_permissions(request, obj)

    def dispatch(self, request, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        request = self.initialize_request(request, *args, **kwargs)
        self.request = request
        self.headers = self.default_response_headers

        try:
            self.initial(request, *args, **kwargs)
            if request.method.lower() in self.http_method_names:
                handler = getattr(self, request.method.lower(),
                                  self.http_method_not_allowed)
            else:
                handler = self.http_method_not_allowed
            self.pre_method_call(request, *args, **kwargs)
            response = handler(request, *args, **kwargs)
            self.post_method_call(request, response, *args, **kwargs)

        except Exception as exc:
            response = self.handle_exception(exc)

        self.response = self.finalize_response(request, response, *args, **kwargs)

        return self.response

    def handle_exception(self, exc):
        """
        high jacking exception and handle with default_exception_handler
        """
        self.pre_handle_exception(exc)
        djackal_handler = self.get_default_exception_handler()
        context = self.get_exception_handler_context()
        response = djackal_handler(exc, context)
        if response is not None:
            response.exception = True
            return response
        else:
            return super().handle_exception(exc)

    def has_auth(self):
        return self.request.user is not None and self.request.user.is_authenticated

    def get_meta(self, **kwargs):
        return kwargs

    def simple_response(self, result=None, status=200, meta=None, headers=None, **kwargs):
        response_data = {}

        if self.result_root:
            response_data[self.result_root] = result
            if self.result_meta:
                meta = self.get_meta(**(meta or dict()))
                response_data[self.result_meta] = meta
        else:
            response_data = result or dict()

        return Response(response_data, status=status, headers=headers, **kwargs)


class DataPurifyMixin:
    data_schema = None

    def get_purified_data(self, key=None, many=False):
        schema = self.get_data_schema(key)
        return purify(self.request.data, schema, many=many)

    def get_data_schema(self, key=None):
        if key in self.data_schema:
            return self.data_schema[key]
        return self.data_schema


class DjackalAPIView(BaseDjackalAPIView, FilterMixin, PageMixin, DataPurifyMixin):
    model = None
    queryset = None

    bind_kwargs_map = {}

    serializer_class = None

    def get_queryset(self):
        assert self.queryset is not None or self.model is not None, (
            '{} should include a `queryset` or `model` attribute'
            'or override the get_queryset() method'.format(self.__class__.__name__)
        )
        if self.queryset is not None:
            queryset = self.queryset.all()
            return queryset
        else:
            return self.model.objects.all()

    def get_model(self):
        if self.model is not None:
            return self.model
        queryset = self.get_queryset()
        assert queryset is not None, (
            '{} should include a `model` or `queryset` attribute'
            'or override the get_model() method'.format(self.__class__.__name__)
        )
        return queryset.model

    def get_request_data(self, key=None):
        return self.get_purified_data(key)

    def get_serializer_class(self):
        return self.serializer_class

    def get_serializer_context(self, **kwargs):
        return kwargs

    def get_serializer(self, instance, context=None, many=False, klass=None):
        if klass is None:
            klass = self.get_serializer_class()
        context = self.get_serializer_context(**(context or dict()))
        ser = klass(instance, many=many, context=context)
        return ser

    def get_bind_kwargs_map(self, **additional):
        d = self.bind_kwargs_map or dict()
        return {**d, **additional}

    def get_bind_kwargs_data(self):
        return value_mapper(self.get_bind_kwargs_map(), self.kwargs)



from djackal.tests import DjackalTestCase
from djackal.utils import isiter, value_mapper


class TestLoader(DjackalTestCase):
    def test_value_mapper(self):
        a_dict = {
            'key1': 'a_value1',
            'key2': 'a_value2',
        }

        b_dict = {
            'key1': 'b_value1',
            'key2': 'b_value2'
        }

        result = value_mapper(a_dict, b_dict)
        assert result[a_dict['key1']] == b_dict['key1']
        assert result[a_dict['key2']] == b_dict['key2']

    def test_iterable(self):
        self.assertTrue(isiter([1, 2, 3]))
        self.assertTrue(isiter((1, 2, 3)))
        self.assertTrue(isiter({1, 2, 3}))
        self.assertTrue(isiter({1: 1, 2: 2, 3: 3}))

        self.assertFalse(isiter('String Sentence'))
        self.assertFalse(isiter(None))
        self.assertFalse(isiter(False))
        self.assertFalse(isiter(True))
        self.assertFalse(isiter(123))



from django.test import TestCase, override_settings

from djackal.filters import DefaultFilterFunc
from djackal.settings import DjackalSettings, djackal_settings


class TestSettings(TestCase):
    def test_settings_attr(self):
        assert djackal_settings.PAGE_SIZE == 10

    def test_import_error(self):
        settings = DjackalSettings({
            'PARAM_FUNC_CLASSES': [
                'tests.invalid.InvalidParamFunc'
            ]
        })
        with self.assertRaises(ImportError):
            print(settings.PARAM_FUNC_CLASSES)

    def test_override_settings(self):
        with override_settings(DJACKAL={'PAGE_SIZE': 20}):
            assert djackal_settings.PAGE_SIZE == 20

        assert djackal_settings.PAGE_SIZE is 10

    def test_str_import(self):
        dq = djackal_settings.PARAM_FUNC_CLASSES[0]
        assert dq is DefaultFilterFunc



from django.db.models import Sum
from django.db.models.functions import Coalesce


class DefaultSum(Coalesce):
    def __init__(self, field, default=0, filter=None):
        super().__init__(Sum(field, filter=filter), default)



from django.db import models

from djackal.fields import JSONField
from djackal.model_mixins.extra_mixin import ExtraMixin
from djackal.tests import DjackalTransactionTestCase


class TestModel(ExtraMixin, models.Model):
    extra_fields = ('b_field1', 'b_field2')
    extra = JSONField(default=dict)
    field_char = models.CharField(max_length=150, null=True)


class TestModel2(ExtraMixin, models.Model):
    extra_field_name = 'b_field'
    extra_fields = ('b_field1', 'b_field2')
    b_field = JSONField(default=dict)


class TestModel3(ExtraMixin, models.Model):
    extra_fields = {
        'b_field1': 'DEFAULT_VALUE',
        'b_field2': list,
    }
    extra = JSONField(default=dict)


class BindMixinTest(DjackalTransactionTestCase):
    def test_bind_values(self):
        tobj = TestModel()
        tobj.b_field1 = 'test_b_field1'
        tobj.field_char = 'char_field'
        tobj.save()
        tobj = TestModel.objects.get(id=tobj.id)
        self.assertEqual(tobj.b_field1, 'test_b_field1')
        self.assertEqual(tobj.extra, {'b_field1': 'test_b_field1'})
        self.assertIsNone(tobj.b_field2)
        self.assertEqual(tobj.field_char, 'char_field')
        with self.assertRaises(AttributeError):
            tobj.b_field3

        tobj.b_field2 = 'test_b_field2'
        tobj.save()
        tobj = TestModel.objects.get(id=tobj.id)
        self.assertEqual(tobj.b_field2, 'test_b_field2')
        self.assertEqual(tobj.extra, {'b_field1': 'test_b_field1', 'b_field2': 'test_b_field2'})

    def test_different_bind_field_name(self):
        tobj = TestModel2()
        tobj.b_field1 = 'test_b_field1'
        tobj.save()
        tobj = TestModel2.objects.get(id=tobj.id)
        self.assertEqual(tobj.b_field1, 'test_b_field1')
        self.assertEqual(tobj.b_field, {'b_field1': 'test_b_field1'})
        self.assertIsNone(tobj.b_field2)
        with self.assertRaises(AttributeError):
            tobj.b_field3

    def test_dictionary_bound_fields(self):
        tobj = TestModel3.objects.create()

        self.assertEqual(tobj.b_field1, tobj.extra_fields['b_field1'])
        self.assertIs(type(tobj.b_field2), tobj.extra_fields['b_field2'])

        changed_value = 'CHANGED_VALUE'
        tobj.b_field1 = changed_value
        tobj.save()
        tobj.refresh_from_db()

        self.assertEqual(changed_value, tobj.b_field1)



from django.db import models

from djackal.shortcuts import get_object_or


class Storage(models.Model):
    key = models.CharField(primary_key=True, max_length=150)
    value = models.TextField()

    @classmethod
    def set(cls, key, value):
        val = cls.get(key)
        if val is None:
            val = cls(key=key)
        val.value = value
        val.save()
        return val

    @classmethod
    def get(cls, key, default=None, f=None):
        val = get_object_or(cls, this=default, key=key)
        if f:
            val = f(val)
        return val

    @classmethod
    def remove(cls, key):
        cls.objects.filter(key=key).delete()




import random

from django.db.models import Q

from djackal.exceptions import NotFound
from djackal.filters import DjackalQueryFilter
from djackal.tests import DjackalTransactionTestCase
from tests.models import TestModel


class TestQueryFilter(DjackalTransactionTestCase):
    def setUp(self):
        self.max_length = 10
        self.objs = [TestModel(
            field_int=i, field_char=random.choice(['A Char', 'B Char', 'C Char']),
            field_text=random.choice(['A Text', 'B Text', 'C Text']),
        ) for i in range(1, self.max_length + 1)]
        TestModel.objects.bulk_create(self.objs)
        self.filter_map = {
            'int_gte': 'field_int__gte',
            'int_lte': 'field_int__lte',
            'field_char': 'field_char__contains',
            'field_text': 'field_text__contains',
            'field_a': 'field_a',
            'field_b': 'field_b'
        }
        self.search_dict = {
            'all': ('field_char__contains', 'field_text__contains', 'field_int__contains'),
            'field_char': 'field_char__contains',
            'field_text': 'field_text__contains',
            'field_int': 'field_int__contains',
        }

    def assertEqualQuerySet(self, set1, set2):
        listed_1 = list(set1.order_by('id').values_list('id', flat=True))
        listed_2 = list(set2.order_by('id').values_list('id', flat=True))
        self.assertEqual(listed_1, listed_2)

    def test_filter_map(self):
        f = DjackalQueryFilter(TestModel.objects.all(), {})
        queryset = f.filter_map(self.filter_map).queryset
        self.assertEqualQuerySet(TestModel.objects.filter(), queryset)

        params = {'int_gte': 5}
        f = DjackalQueryFilter(TestModel.objects.all(), params)
        queryset = f.filter_map(self.filter_map).queryset
        self.assertEqualQuerySet(TestModel.objects.filter(field_int__gte=5), queryset)

        params.update({'field_char': 'A Char'})
        f = DjackalQueryFilter(TestModel.objects.all(), params)
        queryset = f.filter_map(self.filter_map).queryset
        self.assertEqualQuerySet(TestModel.objects.filter(field_int__gte=5, field_char__contains='A Char'), queryset)

    def test_search(self):
        f = DjackalQueryFilter(TestModel.objects.all(), {})
        queryset = f.search(self.search_dict).queryset
        self.assertEqualQuerySet(TestModel.objects.all(), queryset)

        first_model = TestModel.objects.first()
        search_keyword = first_model.field_char

        params = {'search_keyword': search_keyword}
        f = DjackalQueryFilter(TestModel.objects.all(), params)
        queryset = f.search(self.search_dict).queryset
        self.assertEqualQuerySet(TestModel.objects.filter(
            Q(field_char__contains=search_keyword) |
            Q(field_text__contains=search_keyword) |
            Q(field_int__contains=search_keyword)
        ), queryset)
        self.assertIn(first_model, queryset)

        params = {'search_keyword': search_keyword, 'search_type': 'field_text'}
        f = DjackalQueryFilter(TestModel.objects.all(), params)
        queryset = f.search(self.search_dict).queryset
        self.assertEqualQuerySet(TestModel.objects.filter(field_text__contains=search_keyword), queryset)
        self.assertNotIn(first_model, queryset)

    def test_get(self):
        f = DjackalQueryFilter(TestModel.objects.all())
        with self.assertRaises(NotFound):
            f.get(True, field_a=1)

        obj = f.get(field_int=1)
        self.assertIsInstance(obj, TestModel)
        self.assertEqual(obj.field_int, 1)

    def test_extra(self):
        f = DjackalQueryFilter(TestModel.objects.all())
        queryset = f.extra(field_int__gte=5).queryset

        self.assertEqualQuerySet(TestModel.objects.filter(field_int__gte=5), queryset)

        for i, o in enumerate(TestModel.objects.all()):
            o.field_a = self.max_length - i
            o.save()

        f = DjackalQueryFilter(TestModel.objects.all(), {'ordering': '-field_a'})
        queryset = f.ordering().queryset

        for i, o in enumerate(queryset):
            self.assertEqual(o.field_a, self.max_length - i)



DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    },
}

INSTALLED_APPS = (
    'django.contrib.contenttypes',
    'django.contrib.staticfiles',
    'django.contrib.auth',
    'rest_framework',
    'djackal.storage',
    'djackal',
    'tests',
)

MIDDLEWARE = []

USE_TZ = True

SECRET_KEY = 'foobar'

TEMPLATES = [{
    'BACKEND': 'django.template.backends.django.DjangoTemplates',
    'APP_DIRS': True,
}]

STATIC_URL = '/static/'



import warnings

from djackal.shortcuts import get_model


class Initializer:
    def __init__(self):
        self.init_methods = {}

    def wrapper(self, func, key, finish_message=None, skip_message=None, bind_self=False, **options):
        def inner():
            djadis_model = get_model('storage.Storage')
            if not djadis_model.get('{}_init'.format(key), False):
                result = func(self) if bind_self else func()
                djadis_model.set('{}_init'.format(key), True)
                print(finish_message or 'Finished {} initializing'.format(key))
                return result
            else:
                print(skip_message or 'Skip {} initializing'.format(key))

        return inner

    def add(self, key, **options):
        def decorator(func):
            if key in self.init_methods:
                warnings.warn('Duplicated init_method key: {}. It will overwritten by method that defined later.')
            self.init_methods[key] = self.wrapper(func, key, **options)
            return func

        return decorator

    def run(self):
        for key, method in self.init_methods.items():
            method()



import random

from djackal.erra import Erra
from djackal.tests import DjackalTestCase

CONTEXT_MESSAGE = 'Test Erra {num}'


class test(Erra):
    ERRA_ONE = 'Test Erra one'
    ERRA_TWO = 'Test Erra two'
    ERRA_CONTEXT = CONTEXT_MESSAGE


class ErraTest(DjackalTestCase):
    def test_get_message(self):
        random_int = random.randint(1, 100)
        message = test.ERRA_CONTEXT.get_message({'num': random_int})
        self.assertEqual(CONTEXT_MESSAGE.format(num=random_int), message)

    def test_response_data(self):
        response_data = test.ERRA_TWO.response_data()
        self.assertEqual(response_data, {'code': test.ERRA_TWO.code, 'message': test.ERRA_TWO.message})

    def test_str(self):
        self.assertEqual(str(test.ERRA_ONE), 'ERRA_ONE')



from rest_framework.exceptions import APIException


class DjackalAPIException(APIException):
    default_message = ''
    default_status_code = 500

    def __init__(self, message=None, status_code=None):
        self.message = message
        self.status_code = status_code or self.default_status_code

    def __str__(self):
        return self.message

    def response_data(self):
        return {}

    def get_status_code(self):
        return getattr(self, 'status_code', self.default_status_code)


class ErraException(DjackalAPIException):
    default_status_code = 500

    def __init__(self, erra=None, code=None, message=None, context=None, status_code=None, **kwargs):
        self.erra = erra
        self.message = message
        self.code = code
        self.context = context
        self.kwargs = kwargs
        self.status_code = status_code or self.default_status_code

    def __str__(self):
        return self.get_message()

    def get_message(self):
        if self.message:
            return self.message
        elif self.erra:
            return self.erra.message
        return None

    def response_data(self):
        if self.erra:
            response = self.erra.response_data(context=self.context)
        else:
            response = {'code': self.code, 'message': self.message}
        response.update(self.kwargs)
        return response


class NotFound(DjackalAPIException):
    default_status_code = 404
    default_message = 'Data not found'

    def __init__(self, message=None, context=None, model=None, status_code=None, **kwargs):
        self.context = context
        self.kwargs = kwargs
        self.model = model
        self.message = message or self.default_message
        self.status_code = status_code or self.default_status_code

    def response_data(self):
        return {
            'code': 'NOT_FOUND',
            'message': self.message,
            **self.kwargs
        }


class BadRequest(ErraException):
    default_status_code = 400


class Unauthorized(ErraException):
    default_status_code = 401


class Forbidden(ErraException):
    default_status_code = 403


class NotAllowed(ErraException):
    default_status_code = 405


class InternalServer(ErraException):
    default_status_code = 500


class PermissionException(ErraException):
    default_status_code = 403

    def __init__(self, permission, erra=None, code=None, message=None, context=None, status_code=None, **kwargs):
        super().__init__(erra=erra, code=code, message=message, context=context, status_code=status_code, **kwargs)
        self.permission = permission



from rest_framework.response import Response
from rest_framework.views import exception_handler as drf_exception_handler

from djackal.exceptions import DjackalAPIException


def exception_handler(exc, context):
    if isinstance(exc, DjackalAPIException):
        return Response(exc.response_data(), status=exc.get_status_code())
    return drf_exception_handler(exc, context)



from django.test import override_settings

from djackal.tests import DjackalTestCase


class TestLoader(DjackalTestCase):
    def test_initializer_loader(self):
        with override_settings(DJACKAL={
            # 'INITIALIZER':
        }):
            pass



from djackal.shortcuts import model_update

__all__ = [
    'ListViewMixin',
    'CreateViewMixin',
    'DetailViewMixin',
    'UpdateViewMixin',
    'DestroyViewMixin',
]


class ListViewMixin:
    def list(self, request, **kwargs):
        filtered_queryset = self.get_filtered_queryset()

        if self.paging:
            paginate_queryset = self.get_paginate_queryset(filtered_queryset)
            ser = self.get_serializer(paginate_queryset, many=True)
            meta = self.get_paginated_meta()
            return self.simple_response(ser.data, meta=meta)

        ser = self.get_serializer(filtered_queryset, many=True)
        return self.simple_response(ser.data)


class CreateViewMixin:
    def get_create_data(self):
        data = self.get_request_data()
        data.update(**self.get_bind_kwargs_data())
        if self.bind_user_field:
            data[self.bind_user_field] = self.request.user
        return data

    def create_action(self, data):
        model = self.get_model()
        return model.objects.create(**data)

    def create(self, request, **kwargs):
        create_data = self.get_create_data()
        obj = self.create_action(create_data)
        return self.simple_response({'id': obj.id})


class DetailViewMixin:
    def detail(self, request, **kwargs):
        obj = self.get_object()
        ser = self.get_serializer(obj)
        return self.simple_response(ser.data)


class UpdateViewMixin:
    def get_update_data(self):
        data = self.get_request_data()
        data.update(**self.get_bind_kwargs_data())
        if self.bind_user_field:
            data[self.bind_user_field] = self.request.user
        return data

    def update_action(self, obj, data):
        model_update(obj, **data)
        return obj

    def update(self, request, **kwargs):
        obj = self.get_object()
        update_data = self.get_update_data()
        self.update_action(obj, data=update_data)
        return self.simple_response({'id': obj.id})


class DestroyViewMixin:
    def delete_action(self, obj):
        obj.delete()

    def destroy(self, request, **kwargs):
        obj = self.get_object()
        self.delete_action(obj)
        return self.simple_response()



from django.test import override_settings

from djackal.initializer import Initializer
from djackal.loaders import initializer_loader
from djackal.storage.models import Storage
from djackal.tests import DjackalTransactionTestCase

initializer = Initializer()


@initializer.add('test')
def test_init_method():
    return 'test init result'


@initializer.add('test2')
def test_init_method():
    return 'test init result2'


class InitializerTest(DjackalTransactionTestCase):
    def test_initializer(self):
        assert len(initializer.init_methods) == 2
        assert initializer.init_methods['test']() == 'test init result'
        assert initializer.init_methods['test2']() == 'test init result2'
        assert Storage.get('test_init') is not None
        assert Storage.get('test2_init') is not None

    def test_initializer_load(self):
        with override_settings(DJACKAL={'INITIALIZER': 'tests.test_initializer.initializer'}):
            assert initializer_loader() is initializer

    def test_initializer_run(self):
        initializer.run()
        assert Storage.get('test_init') is not None
        assert Storage.get('test2_init') is not None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['djackal',
 'djackal.fields',
 'djackal.management',
 'djackal.management.commands',
 'djackal.model_mixins',
 'djackal.storage',
 'djackal.storage.migrations',
 'djackal.views']

package_data = \
{'': ['*']}

install_requires = \
['djangorestframework>=3.0.0,<4.0.0', 'puty>=0.0,<0.1']

setup_kwargs = {
    'name': 'djackal',
    'version': '0.4.0',
    'description': 'extension of Django REST Framework',
    'long_description': '# Djackal, Django Rest Framework extension\n\n[![version badge](https://badge.fury.io/py/djackal.svg)](https://badge.fury.io/py/djackal)\n\n![djackal image](https://imgur.com/XnlU8T9.jpg)\n\n**Djackal** is extension of Django REST Framework(DRF)\nthat help you easily implement the necessary features on your web backend server.\n\n\n## Warning\n\nThis repository is on developing. Some bugs may exist, possible to change suddenly. And no document yet.\n\n\n## Installation\n\n    pip install djackal\n',
    'author': 'jrog',
    'author_email': 'jrog612@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/jrog612/djackal',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from django.core.management import BaseCommand

from djackal.loaders import initializer_loader


class Command(BaseCommand):
    def handle(self, *args, **options):
        initializer = initializer_loader()
        if not initializer:
            print('No initializers')

        print('Start initializing')
        initializer.run()
        print('Initialing done.')



class ExtraMixin:
    extra_fields = ()
    extra_field_name = 'extra'

    def __init__(self, *args, **kwargs):
        self.extra_field_keys = tuple(self.extra_fields)
        super().__init__(*args, **kwargs)

    def __setattr__(self, key, value):
        if key in ('extra_fields', 'extra_field_name', 'extra_field_keys'):
            return super().__setattr__(key, value)
        if key in self.extra_field_keys:
            extra_hash = getattr(self, self.extra_field_name) or dict()
            extra_hash[key] = value
            return setattr(self, self.extra_field_name, extra_hash)
        return super().__setattr__(key, value)

    def __getattribute__(self, item):
        if item in ('extra_fields', 'extra_field_name', 'extra_field_keys'):
            return super().__getattribute__(item)
        elif item in self.extra_field_keys:
            extra_hash = getattr(self, self.extra_field_name) or dict()
            value = extra_hash.get(item)
            if value is None and type(self.extra_fields) is dict:
                default = self.extra_fields[item]
                value = default() if callable(default) else default
            return value

        return super().__getattribute__(item)



import random

from django.test import override_settings

from djackal.exceptions import NotFound
from djackal.shortcuts import get_object_or_404, get_object_or_None, model_update, get_object_or, get_model, auto_f_key, \
    gen_q
from djackal.tests import DjackalTransactionTestCase
from tests.models import TestModel


class TestShortcuts(DjackalTransactionTestCase):
    def test_get_object_or(self):
        obj = TestModel.objects.create(field_int=1)

        self.assertIsNone(get_object_or_None(TestModel, field_int=2))
        self.assertEqual(get_object_or_None(TestModel, field_int=1), obj)

        with self.assertRaises(NotFound) as res:
            get_object_or_404(TestModel, field_int=2)

        self.assertIs(res.exception.model, TestModel)
        self.assertEqual(get_object_or_404(TestModel, field_int=1), obj)

        self.assertEqual('TestModel', get_object_or(TestModel, 'TestModel', field_int=2))
        self.assertEqual(obj, get_object_or(TestModel, 'TestModel', field_int=1))

    def test_model_update(self):
        obj = TestModel.objects.create(field_int=1, field_char='text')

        obj = model_update(obj, field_int=2, field_char='test2')
        obj.refresh_from_db()

        self.assertEqual(obj.field_int, 2)
        self.assertEqual(obj.field_char, 'test2')

        obj = model_update(obj, field_int=4, field_char='test4', commit=False)
        obj.refresh_from_db()

        self.assertEqual(obj.field_int, 2)
        self.assertEqual(obj.field_char, 'test2')

    def test_get_model(self):
        model = get_model('tests.TestModel')
        self.assertEqual(TestModel, model)

        with self.assertRaises(ValueError):
            get_model('TestModel')

        with override_settings(DJACKAL={
            'SINGLE_APP': True,
            'SINGLE_APP_NAME': 'tests'
        }):
            model = get_model('TestModel')
            self.assertEqual(TestModel, model)

    def test_gen_q(self):
        random_int = random.randint(1, 100)
        random_char = f'random_char_{random_int}'
        tm1 = TestModel.objects.create(field_int=random_int)
        tm2 = TestModel.objects.create(field_char=random_char)
        TestModel.objects.create(field_char='WRONG')
        TestModel.objects.create(field_int=random_int + 100)

        qs = gen_q(random_int, 'field_int', 'field_char__contains')
        result = TestModel.objects.filter(qs)
        self.assertIn(tm1, result)
        self.assertIn(tm2, result)
        self.assertLen(2, result)

    def test_auto_f_key(self):
        tm = TestModel.objects.create()

        self.assertEqual(auto_f_key(test_model=1), {'test_model_id': 1})
        self.assertEqual(auto_f_key(test_model=tm), {'test_model': tm})

        with self.assertRaises(ValueError):
            auto_f_key(test_model='TestModel'),



from rest_framework import serializers


class BaseModelSerializer(serializers.ModelSerializer):
    extra_standard_fields = ()

    def __new__(cls, *args, **kwargs):
        if kwargs.get('many', False) is True:
            context = kwargs.get('context', {})
            context.update(has_many=kwargs.get('many', True))
            kwargs.update(context=context)
        return super().__new__(cls, *args, **kwargs)

    @property
    def has_many(self):
        return self.context.get('has_many', False)

    def get_fields(self):
        fields = super().get_fields()

        for field_name in fields:
            if isinstance(fields[field_name], serializers.ListSerializer):
                if isinstance(fields[field_name].child, BaseModelSerializer):
                    fields[field_name].child._context = self._context
            elif isinstance(fields[field_name], BaseModelSerializer):
                fields[field_name]._context = self._context

        return fields

    def request(self):
        return self.context.get('request')

    def current_user(self):
        request = self.request()
        if request is None:
            return self.context.get('user') or self.context.get('current_user')
        return request.user

    def build_field(self, field_name, info, model_class, *args, **kwargs):
        model_field = model_class._meta.get_field(field_name)
        if model_field.__class__ in self.extra_standard_fields:
            return self.build_standard_field(field_name, model_field)
        return super().build_field(field_name, info, model_class, *args, **kwargs)



from enum import Enum


class Erra(Enum):
    def __str__(self):
        return self.name

    @property
    def message(self):
        return self.value

    @property
    def code(self):
        return self.name

    def get_message(self, context=None):
        message = self.message
        if context and message:
            message = message.format(**context)
        return message

    def response_data(self, context=None):
        message = self.get_message(context=context)
        return {'code': self.code, 'message': message}



from django.db import models
from rest_framework import serializers


class TestModel(models.Model):
    field_char = models.CharField(null=True, max_length=1)
    field_int = models.IntegerField(null=True)
    field_text = models.TextField(null=True)
    field_a = models.IntegerField(null=True)
    field_b = models.IntegerField(null=True)
    field_bool = models.BooleanField(default=True)


class TestSerializer(serializers.ModelSerializer):
    class Meta:
        model = TestModel
        fields = '__all__'
